from neo4j import GraphDatabase
from statistique import mean

class ClientNeo4j:
    def __init__(self, uri, user, password):
        self.driver = GraphDatabase.driver(uri, auth=(user, password))
        self.password = password
        self.number_process = int(self.make_query("Match(n:process) Return Count(n)")[0][0])

    def close(self):
        self.driver.close()

    def make_query(self, query):
        with self.driver.session() as session:
            result = session.read_transaction(self.generate_query, query)
        return result

    def set_heuristique(self, dict):
        with self.driver.session() as session:
            for threat_value in dict:
                query = f"Match(n:process) Where ID(n) IN {dict[threat_value]} SET n.heuristic = {threat_value[0]} SET n.heuristic_detail = {list(threat_value[1])}"
                session.write_transaction(self.generate_query, query)
        return 1

    def generate_query(self, tx, query):
        output = []
        res = tx.run(query)
        for out in res:
            output.append(out)
        return output

    def get_heuristic(self,list_id):
        res = self.make_query(f"Match (n:process) Where ID(n) IN {list_id} Return AVG(n.heuristic)")
        return res[0][0]

    def get_threats(self):
        with self.driver.session() as session:
            query = f"Match(n:threat)--(p:producer) RETURN n, p.name"
            results = session.read_transaction(self.generate_query, query)
            nodes = []
            for res in results:
                node = dict(res[0])
                node["_id"] = res[0].id
                node["producer"] = res[1]
                node["hf"] = res[0]["heuristic_detail"][0]
                node["df"] = res[0]["heuristic_detail"][1]
                node["pl"] = res[0]["heuristic_detail"][2]
                node["nf"] = res[0]["heuristic_detail"][3]
                node["nc"] = res[0]["heuristic_detail"][4]
                node["ct"] = res[0]["heuristic_detail"][5]
                node["nd"] = res[0]["heuristic_detail"][6]
                nodes.append(node)
            return nodes

    def mark_as_a_threat(self,id):
        with self.driver.session() as session:
            query = f"Match(n:process) Where ID(n) = {id} SET n:process:threat"
            session.write_transaction(self.generate_query, query)
        pass


    def retrieve_attribut(self, node_type, attribut_name, limit=None):

        query = "MATCH (n:" + node_type + ") RETURN n." + attribut_name
        if limit is not None:
            query += " LIMIT " + str(limit)

        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            result = []
            for res in raw_result:
                res[0]["_id"] = res[0].id
                result.append(res[0])
        return result

    def retrieve_process(self, id, number_of_child=2, number_of_parents=1):
        query_node = f"Match (k:process) WHERE ID(k)={id} return k"
        query_child = f"Match (k:process)-[:father_of*1..{number_of_child}]->(x:process) WHERE ID(k)= {id} return x"
        query_parent = f"Match (x:process)-[:father_of*1..{number_of_parents}]->(k:process) WHERE ID(k)= {id} return x"
        query_hash = f"Match (x:hash)-[:hash_link]->(k:process) WHERE ID(k)= {id} return x"
        query_producer = f"Match (k:process)-[]->(x:producer) WHERE ID(k)= {id} return x"
        query_directory = f"Match (k:process)--(x:path_directory) WHERE ID(k)= {id} return x"
        query_st = f"Match p=(k:process)-[:father_of*1..{number_of_child}]->(x:process) WHERE ID(k)={id} WITH nodes(p) AS n return [node in n | id(node)]"
        query_ts = f"Match p=(x:process)-[:father_of*1..{number_of_parents}]->(k:process) WHERE ID(k)={id} WITH nodes(p) AS n return [node in n | id(node)]"

        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query_node)
            result_node = None
            for res in raw_result:
                result_node = res[0]
                break
            raw_result = session.read_transaction(self.generate_query, query_child)
            result_child = []
            for res in raw_result:
                result_child.append(res[0])
            raw_result = session.read_transaction(self.generate_query, query_parent)
            result_parent = []
            for res in raw_result:
                result_parent.append(res[0])
            raw_result = session.read_transaction(self.generate_query, query_hash)
            result_hash = []
            for res in raw_result:
                result_hash.append(res[0])
            raw_result = session.read_transaction(self.generate_query, query_producer)
            result_producer = []
            for res in raw_result:
                result_producer.append(res[0])
            raw_result = session.read_transaction(self.generate_query, query_directory)
            result_directory = []
            for res in raw_result:
                result_directory.append(res[0])
            target_child = {}
            raw_result = session.read_transaction(self.generate_query, query_st)
            for res in raw_result:
                line = res[0]
                for index in range(len(line)-1):
                    if line[index] in target_child:
                        if line[index+1] not in target_child[line[index]]:
                            target_child[line[index]].append(line[index+1])
                    else:
                        target_child[line[index]] = [line[index+1]]
            target_parent = {}
            raw_result = session.read_transaction(self.generate_query, query_ts)
            for res in raw_result:
                line = res[0]
                for index in range(len(line) - 1):
                    if line[index] in target_parent:
                        if line[index + 1] not in target_parent[line[index]]:
                            target_parent[line[index]].append(line[index + 1])
                    else:
                        target_parent[line[index]] = [line[index + 1]]
            links = []
            for source in target_child :
                for target in target_child[source] :
                    links.append({"source": source, "target": target})
            for source in target_parent :
                for target in target_parent[source] :
                    links.append({"source": source, "target": target})

            return result_node, result_child, result_parent, result_hash, result_producer, result_directory, links

    def search_processes(self, value, by="process_exec"):
        value = value.lower().replace("\\", "\\\\")

        condition = f"LOWER(n.{by}) CONTAINS '{value}' OR LOWER(h.hash) = '{value}'"
        if by in ("pid",) :
            condition = f"n.{by} = '{value}'"

        query = "MATCH (n) " \
                "MATCH (n)--(d:path_directory) " \
                "MATCH (n)--(h:hash) " \
                "MATCH (n)--(p:producer) " \
                f"WHERE {condition} " \
                "RETURN n, ID(n), d.name, p.name, " \
                "h.hash, h.malicious, h.suspicious, h.undetected"
        print(query)
        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            result = []
            for res in raw_result:
                node = {
                    'guid': res[0]["guid"],
                    'pid': res[0]["pid"],
                    'process_path': res[0]["process_path"],
                    'process_exec': res[0]["process_exec"],
                    '_id': int(res[1]),
                    'directory': res[2],
                    'hash': res[4],
                    'malicious': int(res[5]),
                    'suspicious': int(res[6]),
                    'undetected': int(res[7]),
                    'producer': res[3],
                    'threat': res[0]["heuristic"]
                }
                result.append(node)
        return result

    def retrieve_process_by_attribut(self, attribut, value):
        value = value.replace("\\", "\\\\")
        if attribut == "directory":
            query = f'MATCH (n:process)--(d:path_directory) WHERE d.name =\'{value}\' ' \
                    f'RETURN n.process_exec, count(n) ORDER BY count(n) DESC'
        elif attribut == "hash":
            query = f'MATCH (n:process)--(h:hash) WHERE h.hash = \'{value}\' RETURN n.process_exec, COUNT(n) ' \
                    f'ORDER BY count(n) DESC'
        elif attribut == "producer":
            query = f'MATCH (n:process)--(p:producer) WHERE p.name = \'{value}\' RETURN n.process_exec, COUNT(n) ' \
                    f'ORDER BY count(n) DESC '
        elif attribut == "path_length":
            query = f'MATCH (n:process)-[]-(d:path_directory) WITH n AS n, d AS d, SIZE(split(d.name, "\\\\"))-1 AS c ' \
                    f'WHERE c = {value} RETURN d.name, COUNT(n) ORDER BY COUNT(n) DESC'
        else:
            query = f'MATCH (n:process) WHERE n.{attribut} = \'{value}\' RETURN n.process_exec, COUNT(n)'

        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            result = []
            for res in raw_result:
                result.append({"process_exec": res[0], "count": int(res[1])})
        return result

    def build_path_length(self):

        query = "MATCH (n:process)-[]-(d:path_directory) RETURN d, d.name"

        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            result_path_length = {}
            for res in raw_result:
                # exemple : C:\Users\ADMINISTRATOR\Downloads\Sysmon\
                tmp = len(res[1].rstrip('\\').split("\\"))
                if tmp not in result_path_length:
                    result_path_length[tmp] = 1
                else:
                    result_path_length[tmp] += 1
        return result_path_length

    def build_common_directory(self):

        query = "MATCH (d:path_directory)-[]-(n) RETURN d.name, COUNT(d) ORDER BY COUNT(d) DESC"

        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            result_common_directory = {}
            for res in raw_result:
                # exemple : C:\Users\ADMINISTRATOR\Downloads\Sysmon\
                result_common_directory[str(res[0])] = int(res[1])
        return result_common_directory

    def build_frequency(self):

        count_exec = {}
        query = "MATCH (n:process) RETURN n.process_exec ,COUNT(n.process_exec) ORDER BY COUNT(n.process_exec) DESC"
        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            for res in raw_result:
                count_exec[res[0]] = int(res[1])
            return count_exec

    def build_hash(self):
        query = "MATCH (d:hash)-[]-(n) RETURN d, COUNT(d)"

        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            result_hash = {}
            for res in raw_result:
                result_hash[res[0]["hash"]] = int(res[1])
        return result_hash

    def build_producer(self):
        query = "MATCH (d:producer)-[]-(n) RETURN d, COUNT(d)"

        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            result_producer = {}
            for res in raw_result:
                result_producer[res[0]["name"]] = int(res[1])
        return result_producer

    def build_children(self):

        mean_children = {}

        with self.driver.session() as session:

            query = "MATCH p=(n:process)-[f:father_of]->(:process) " \
                    "WITH  n.process_exec AS pe ,n.time AS t,COUNT(p) AS cou " \
                    "RETURN pe ,cou, t ORDER BY cou DESC"
            raw_result = session.read_transaction(self.generate_query, query)

            for res in raw_result:
                key = res[0]
                val = int(res[1])
                if key not in mean_children:
                    mean_children[key] = [val]
                else:
                    mean_children[key].append(val)

        for key in mean_children:
            val = mean_children[key]
            mean_children[key] = mean(val)

        return mean_children

    def build_different_directory(self):
        query = "match (n:process)--(d:path_directory) return n.process_exec, count(distinct d)"

        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            result_diff_directory = {}
            for res in raw_result:
                result_diff_directory[res[0]] = int(res[1])
        return result_diff_directory

    def retrieve_process_in_directory(self, path_directory):
        query = "MATCH (n:process) -[:launch_from]-(d:path_directory{name:'" + path_directory + "'}) " \
                "RETURN n.process_exec, COUNT(n) ORDER BY COUNT(n)"

        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            result_directory = {}
            for res in raw_result:
                result_directory[res[0]] = int(res[1])
        return result_directory

    def find_process(self, string):

        query = "MATCH (n:process) " \
                "WHERE n.process_exec CONTAINS '" + string + "' OR n.process_path CONTAINS '" + string +"'" \
                "RETURN n"

        with self.driver.session() as session:
            raw_result = session.read_transaction(self.generate_query, query)
            result = []
            for res in raw_result:
                result.append(res[0])
        return result
