import collections
import math
from sklearn.ensemble import IsolationForest
from sklearn.cluster import KMeans
import pandas as pd

class ThreatHunter:
    def __init__(self,connector):
        self.con = connector
        self.directory_f = connector.build_common_directory()
        self.path_lenght_l = connector.build_path_length()
        self.path_lenght_f = collections.Counter(self.path_lenght_l)
        self.exec_f = connector.build_frequency()
        self.hash_f = connector.build_hash()
        self.producer_f = connector.build_producer()
        self.build_different_directory_l = connector.build_different_directory()
        self.criteria = {
            "path_length": self.path_lenght_f,
            "directory": self.directory_f,
            "exec": self.exec_f,
            "hash": self.hash_f,
            "producer": self.producer_f
        }

        self.trust_list = {
            "Microsoft Corporation":4,
            "Microsoft Corp.":4,
            "-":0,
            "Martin Prikryl":1,
            "Sysinternals - www.sysinternals.com":3,
            "Apache Software Foundation":2,
            "Splunk Inc.":2,
            "gentilkiwi (Benjamin DELPY)":1,
            "Intezer Labs Ltd.":3,
            "迟思堂工作室":0,
            "Sysinternals":3
        }
        self.max_pl = max(self.path_lenght_f.values())
        self.maxdf = max(self.directory_f.values())
        self.ef = max(self.exec_f.values())
        self.bdd = max(self.build_different_directory_l.values())
        self.clusters = None

    def get_freq(self):
        string = "MATCH (n:process) MATCH (h:hash)-[]->(n:process)-[:launch_from]->(d:path_directory) MATCH (n:process)-[:produced_by]->(c:producer) OPTIONAL MATCH p=(n:process)-[:father_of]->(:process) RETURN n.guid,n.process_path,n.process_exec,COUNT(p),d.name,h.malicious,h.suspicious,h.undetected,c.name,ID(n) ORDER BY COUNT(p) DESC"
        results = self.con.make_query(string)
        freq = []
        for res in results:
            node = {
                'guid': res[0],
                'process_path': res[1],
                'process_exec': res[2],
                'number_child': int(res[3]),
                'dir_name': res[4],
                'malicious': int(res[5]),
                'suspicious': int(res[6]),
                'undetected': int(res[7]),
                'producer_name': res[8],
                '_id': int(res[9]),
                'threat': 0
            }
            hash_f = self.hash_heuristic(node, 1 / 3)[1]
            directory_f = self.directory_heuristic(node, 1)[1]
            path_f = self.path_heuristic(node, 3)[1]
            exec_f = self.exec_heuristic(node, 2)[1]
            child_heuristic = self.child_heuristic(node, 1)[0]
            trust_f = self.trust_factor(node)[1]
            common_dir_f = self.common_dir_heuritic(node, 1)[1]
            freq.append([node['_id'],hash_f,directory_f,path_f,exec_f,child_heuristic,trust_f,common_dir_f])
        return freq



    def big_chungus_query(self):
        string = "MATCH (n:process) MATCH (h:hash)-[]->(n:process)-[:launch_from]->(d:path_directory) MATCH (n:process)-[:produced_by]->(c:producer) OPTIONAL MATCH p=(n:process)-[:father_of]->(:process) RETURN n.guid,n.process_path,n.process_exec,COUNT(p),d.name,h.malicious,h.suspicious,h.undetected,c.name,ID(n) ORDER BY COUNT(p) DESC"
        results = self.con.make_query(string)
        trust_results = []
        dict_threat = {}
        for res in results:
            node = {
                'guid' : res[0],
                'process_path' : res[1],
                'process_exec' : res[2],
                'number_child' : int(res[3]),
                'dir_name' : res[4],
                'malicious' : int(res[5]),
                'suspicious' : int(res[6]),
                'undetected' : int(res[7]),
                'producer' : res[8],
                '_id': int(res[9]),
                'threat' : 0
            }
            temp1,temp2 = self.heuristic(node)
            node['threat'], detail = round(temp1,2),tuple([round(val,2) for val in temp2])
            trust_results.append(node)
            if (node['threat'],detail) in dict_threat:
                dict_threat[(node['threat'],detail)].append(node['_id'])
            else:
                dict_threat[(node['threat'],detail)] = [node['_id']]
        self.con.set_heuristique(dict_threat)
        return sorted(trust_results, key=lambda node: node["threat"], reverse=True)

    def get_results(self, limit=0, force_reload=False):
        if limit <= 0:
            string = "MATCH (n:process) MATCH (h:hash)-[]->(n:process)-[:launch_from]->(d:path_directory) MATCH (n:process)-[:produced_by]->(c:producer) OPTIONAL MATCH p=(n:process)-[:father_of]->(:process) RETURN n.heuristic,n,COUNT(p),d.name,h.malicious,h.suspicious,h.undetected,c.name,ID(n) ORDER BY n.heuristic DESC"
        else:
            string = f"MATCH (n:process) MATCH (h:hash)-[]->(n:process)-[:launch_from]->(d:path_directory) MATCH (n:process)-[:produced_by]->(c:producer) OPTIONAL MATCH p=(n:process)-[:father_of]->(:process) RETURN n.heuristic,n,COUNT(p),d.name,h.malicious,h.suspicious,h.undetected,c.name,ID(n) ORDER BY n.heuristic DESC LIMIT {limit}"
        results = self.con.make_query(string)
        trust_results = []
        data = []
        for res in results:
            node = dict(res[1])
            node.update({
                'guid' : res[1]["guid"],
                'process_path' : res[1]["process_path"],
                'process_exec' : res[1]["process_exec"],
                'number_child' : int(res[2]),
                'dir_name' : res[3],
                'malicious' : int(res[4]),
                'suspicious' : int(res[5]),
                'undetected' : int(res[6]),
                'producer' : res[7],
                '_id': int(res[8]),
                'threat' : float(res[1]["heuristic"]),
                'threat_detail': list(res[1]["heuristic_detail"]),
                "hf": res[1]["heuristic_detail"][0],
                "df": res[1]["heuristic_detail"][1],
                "pl": res[1]["heuristic_detail"][2],
                "nf": res[1]["heuristic_detail"][3],
                "nc": res[1]["heuristic_detail"][4],
                "ct": res[1]["heuristic_detail"][5],
                "nd": res[1]["heuristic_detail"][6]
            })
            trust_results.append(node)
            if (limit == 0):
                """
                KMEANS:
                """
                hash_f = self.hash_heuristic(node, 1 / 3)[1]
                directory_f = self.directory_heuristic(node, 1)[1]
                path_f = self.path_heuristic(node, 3)[1]
                exec_f = self.exec_heuristic(node, 2)[1]
                child_heuristic = self.child_heuristic(node, 1)[0]
                trust_f = self.trust_factor(node)[1]
                common_dir_f = self.common_dir_heuritic(node, 1)[1]
                data.append([node['_id'], hash_f, directory_f, path_f, exec_f, child_heuristic, trust_f, common_dir_f,node])
        if (limit <= 0):
            df = pd.DataFrame(data=data)
            df.columns = ['_id', 'hash_f', 'directory_f', 'path_f', 'exec_f', 'child_heuristic', 'trust_f', 'common_dir_f','nodes']
            df_guid = df['_id'].to_frame()
            df_guid['nodes'] = df['nodes']
            df = df.drop(columns=['_id','nodes'])
            kmeans = KMeans(n_clusters=13, random_state=0).fit(df)
            lab = kmeans.labels_
            df_guid['label'] = lab
            res = df_guid.groupby(['label'])
            threat_dict = {}
            for i in range(13):
                _id = res.get_group(i)['_id'].values.tolist()
                _nodes = res.get_group(i)['nodes'].values.tolist()
                threat_dict[self.con.get_heuristic(_id)] = _nodes
            threat_dict = sorted(threat_dict.items(), key=lambda item: item[0],reverse=True)
            self.clusters = threat_dict
            return sorted(trust_results, key=lambda node: node["threat"], reverse=True),threat_dict
        else:
            return sorted(trust_results, key=lambda node: node["threat"], reverse=True),None


    def heuristic(self,node):
        hash = self.hash_heuristic(node, 1 / 3)[0]
        directory = self.directory_heuristic(node, 1)[0]
        path = self.path_heuristic(node, 3)[0]
        exec = self.exec_heuristic(node, 2)[0]
        child = self.child_heuristic(node, 1)[0]
        trust = self.trust_factor(node)[0]
        common_dir = self.common_dir_heuritic(node,1)[0]
        h  = 2*hash + 2*trust + directory + path + exec + child + trust + common_dir
        return h,(hash,directory,path,exec,child,trust,common_dir)

#"{guid}" {guid}
    def hash_heuristic(self,node,coeff):
        malicious = node['malicious']
        suspicious = node['suspicious']
        undetected = node['undetected']
        sum = malicious + 0.5*suspicious + undetected
        if sum == 0:
            h = 0
            freq = 0
        else:
            freq = (malicious + 0.5 * suspicious) / sum
            h = (freq/sum)**coeff
        return h,freq

    def directory_heuristic(self,node,coeff):
        dir = node['dir_name']
        dir_f = self.directory_f[dir]
        maxi = self.maxdf
        freq = dir_f / maxi
        h = - (freq)**coeff + 1

        return h,freq

    def path_heuristic(self,node,coeff):
        path = node['process_path']
        path_len = len(path.rstrip('\\').split("\\"))
        path_len_f = self.path_lenght_f[path_len]
        maxi = self.max_pl
        freq = path_len_f/maxi
        h = - (freq) ** coeff + 1
        return h,freq


    def exec_heuristic(self,node,coeff):
        exec_name = node['process_exec']
        exec_name_f = self.exec_f[exec_name]
        maxi = self.ef
        freq = exec_name_f/maxi
        h = - (freq) ** coeff + 1
        return h,freq


    def child_heuristic(self,node,coeff):
        number_child = node['number_child']
        freq = number_child/coeff
        h = math.exp(-freq)
        return h,freq

    def trust_factor(self,node):
        producter = node['producer']
        h = abs(self.trust_list[producter] - 4)/4
        freq = h
        return h,freq

    def common_dir_heuritic(self,node,coeff):
        process_exe = node['process_exec']
        num_dir = self.build_different_directory_l[process_exe]
        maxi = self.bdd
        freq = num_dir / maxi
        h =  (freq) ** coeff
        return h,freq
